import math
import os
import shutil
from pathlib import Path

import altair as alt
import pandas as pd
import pint
import seaborn as sns
import yaml

from pkdb_analysis.filter import f_dosing_in, f_mt_in_substance_in
from pkdb_analysis.meta_analysis import MetaAnalysis


alt.data_transformers.disable_max_rows()
# alt.data_transformers.enable('json')

ureg = pint.UnitRegistry()


class LegendArgs(object):
    def __init__(self, field, init=None):
        self.field = field
        self.init = init


class PlottingParameter(object):
    def __init__(
        self,
        measurement_types,
        units_rm,
        infer_by_intervention=True,
        infer_by_output=True,
    ):
        self.measurement_types = measurement_types
        self.units_rm = units_rm
        self.infer_by_intervention = infer_by_intervention
        self.infer_by_output = infer_by_output

    @property
    def joined_measurement_type(self):
        return "_".join(self.measurement_types)


def column_to_color(column):
    mapping = {v: n for n, v in enumerate(column.unique())}
    colors = sns.color_palette(
        "colorblind",
        8,
    ).as_hex()

    return column.apply(lambda x: colors[mapping[x]])


def expand_df(df, on):
    result_df = pd.DataFrame()
    black = "#323232"
    for choice, mc in on.items():
        this_df = df.copy()

        this_df["show"] = choice
        this_df["show_value"] = df[mc.field]

        for key in on.keys():
            this_df[f"color_{key}"] = black

        this_df[f"color_{choice}"] = column_to_color(df[mc.field])
        this_df[f"color"] = this_df[f"color_{choice}"]
        result_df = result_df.append(this_df)
    return result_df.reset_index(drop=True)


def abs_abs(d):
    return (d["per_bw"] == False) & (d["intervention_per_bw"] == False)


def legend(df, selection, title, value_field, color_field=None, color=None):
    if color_field is not None:
        color = alt.Color(color_field, scale=None, legend=None)

    return (
        alt.Chart(df)
        .mark_rect()
        .encode(
            y=alt.Y(value_field, axis=alt.Axis(title=None)),
            color=alt.condition(selection, color, alt.value("white"), legend=None),
            size=alt.value(250),
        )
        .properties(selection=selection, title=title)
    )


def round_up(n, decimals=0):
    multiplier = 10 ** decimals
    return math.ceil(n * multiplier) / multiplier


def create_interactive_plot(
    df, path, tooltip, multi_legend, multi_color_legend, create_json=True
):
    df = expand_df(df, multi_color_legend)
    measurement_type = df["measurement_type"].unique()[0]
    u_unit = ureg(df["unit"].unique()[0])
    u_unit_intervention = ureg(df["intervention_unit"].unique()[0])
    substance = df["substance"].unique()[0]
    substance_intervention = df["intervention_substance"].unique()[0]

    y_axis_label = f"{substance} {measurement_type}"
    y_title = f"{y_axis_label} [{u_unit.u :~P}]".capitalize()
    x_label = f"{substance_intervention} dose"
    x_title = f"{x_label} [{u_unit_intervention.u :~P}]".capitalize()

    y_domain = alt.selection_interval(bind="scales", encodings=["y"])
    x_domain = alt.selection_interval(bind="scales", encodings=["x"])
    brush = alt.selection(type="interval")

    input_dropdown = alt.binding_select(options=df["show"].unique().tolist(), name=" ")
    selection_base = alt.selection_single(
        fields=["show"], bind=input_dropdown, init={"show": "Sex"}
    )

    multi_color_legend_fields = {
        key: mc.field for key, mc in multi_color_legend.items()
    }
    multi_legends = {**multi_color_legend_fields, **multi_legend}
    selections_multi_color = {}

    for key, mc in multi_color_legend.items():
        kwargs = {"fields": [mc.field]}
        if mc.init is not None:
            kwargs["init"] = [{mc.field: field for field in mc.init}]
        selections_multi_color[key] = alt.selection_multi(**kwargs)

    selections_multi = {
        key: alt.selection_multi(fields=[value]) for key, value in multi_legend.items()
    }
    selections_multi = {**selections_multi, **selections_multi_color}

    def filter_transform_no_brush(base):
        base = base.transform_filter(selection_base)
        for key, selection in selections_multi.items():
            base = base.transform_filter(selection)
        return base

    def filter_transform_all(base):
        return filter_transform_no_brush(base).transform_filter(brush)

    y_max = df.y.max()
    y_min = df.y.min()
    x_max = df.intervention_value.max()
    x_min = df.intervention_value.min()

    size_max = df.subject_count.max()
    bins = 50

    stepsize_y = (y_max - y_min) / bins
    stepsize_x = (x_max - x_min) / bins

    titles = {}

    for key in ["age", "weight"]:
        try:
            _u_unit = ureg(df[f"unit_{key}"].dropna().unique()[0])
            titles[key] = f"{key} [{_u_unit.u :~P}]".capitalize()
        except:
            # FIXME: too broad except
            titles[key] = f'{key} [""]'.capitalize()

    relevant_columns = {
        "age",
        "min_sd_age",
        "max_sd_age",
        "weight",
        "min_sd_weight",
        "max_sd_weight",
        "intervention_value",
        "y",
        "y_min",
        "y_max",
        "show",
        "show_value",
        "color",
        "data_type",
        "url",
        "subject_count",
    }

    color_columns = {f"color_{name}" for name in multi_color_legend.keys()}
    tooltip_columns = {name[:-2] for name in tooltip}

    relevant_columns = relevant_columns.union(set(multi_legends.values())).union(
        color_columns
    )
    relevant_columns = relevant_columns.union(tooltip_columns)

    # FIXME (? what does this FIXME mean?)
    df = df[relevant_columns.intersection(df.columns)]

    x_age = alt.X(
        "age:Q",
        scale=alt.Scale(domain=[0, 100], clamp=True),
        axis=alt.Axis(title=titles["age"]),
    )
    y_weight = alt.Y(
        "weight:Q",
        scale=alt.Scale(domain=[0, 150], clamp=True),
        axis=alt.Axis(title=titles["weight"]),
    )

    errorbars_y = filter_transform_no_brush(
        alt.Chart(df)
        .mark_errorbar()
        .encode(
            x=alt.X(
                "intervention_value:Q",
                axis=alt.Axis(title=x_title),
                scale=alt.Scale(domain=x_domain, clamp=True),
            ),
            y=alt.Y(
                "y_min:Q",
                axis=alt.Axis(title=y_title),
                scale=alt.Scale(domain=y_domain, clamp=True),
            ),
            y2="y_max:Q",
            color=alt.Color("color:N", scale=None, legend=None),
            opacity=alt.condition(brush, alt.OpacityValue(1), alt.OpacityValue(0.1)),
        )
    ).add_selection(
        brush
    )  # .add_selection(selection_base)

    errorbars_weight = (
        filter_transform_no_brush(
            alt.Chart(df)
            .mark_errorbar()
            .encode(
                x=x_age,
                y=alt.Y(
                    "min_sd_weight:Q",
                    scale=alt.Scale(domain=[0, 150], clamp=True),
                    axis=alt.Axis(title=titles["weight"]),
                ),
                y2="max_sd_weight:Q",
                color=alt.Color("color:N", scale=None, legend=None),
                opacity=alt.condition(
                    brush, alt.OpacityValue(1), alt.OpacityValue(0.1)
                ),
            )
        )
        .properties(height=300, width=500)
        .add_selection(brush)
        .add_selection(selection_base)
    )

    errorbars_age = (
        filter_transform_no_brush(
            alt.Chart(df)
            .mark_errorbar()
            .encode(
                x=alt.X(
                    "min_sd_age:Q",
                    scale=alt.Scale(domain=[0, 100], clamp=True),
                    axis=alt.Axis(title=titles["age"]),
                ),
                x2="max_sd_age:Q",
                y=y_weight,
                color=alt.Color("color:N", scale=None, legend=None),
                opacity=alt.condition(
                    brush, alt.OpacityValue(1), alt.OpacityValue(0.1)
                ),
            )
        )
        .properties(height=300, width=500)
        .add_selection(brush)
        .add_selection(selection_base)
    )

    domain_marker = ["publication", "from timecourse", "from bodyweight"]
    range_marker = ["circle", "square", "triangle"]

    scatter = filter_transform_no_brush(
        alt.Chart(df)
        .properties(height=350, width=500)
        .mark_point(opacity=0.8, filled=True)
        .encode(
            x=alt.X(
                "intervention_value:Q",
                axis=alt.Axis(title=x_title),
                scale=alt.Scale(domain=x_domain, clamp=True),
            ),
            y=alt.Y(
                "y:Q",
                axis=alt.Axis(title=y_title),
                scale=alt.Scale(domain=y_domain, clamp=True),
            ),
            color=alt.Color("color:N", scale=None, legend=None),
            opacity=alt.condition(brush, alt.OpacityValue(1), alt.OpacityValue(0.1)),
            shape=alt.Shape(
                "data_type:N",
                scale=alt.Scale(domain=domain_marker, range=range_marker),
                legend=alt.Legend(title="Data type", orient="left"),
            ),
            size=alt.Size(
                "subject_count:Q",
                scale=alt.Scale(domain=[0, size_max]),
                legend=alt.Legend(title="Number of subjects", orient="left"),
            ),
            href="url:N",
            tooltip=tooltip,
        )
        .add_selection(brush)
        .add_selection(selection_base)
    )

    bars = filter_transform_all(
        alt.Chart(df)
        .properties(width=500)
        .mark_bar()
        .encode(
            y=alt.Y("show_value:N", axis=alt.Axis(title=None)),
            color=alt.Color("color:N", scale=None, legend=None),
            x=alt.X("sum(subject_count):Q", axis=alt.Axis(title="Number of subjects")),
        )
    )

    histo_y = filter_transform_all(
        alt.Chart(df)
        .properties(height=350, width=80)
        .mark_bar(
            opacity=0.9,
        )
        .encode(
            y=alt.Y(
                "y:Q",
                axis=None,
                bin=alt.Bin(maxbins=bins, minstep=stepsize_y),
                scale=alt.Scale(domain=[y_min, y_max], clamp=True),
            ),
            color=alt.Color("color:N", scale=None, legend=None),
            x=alt.X("sum(subject_count):Q", axis=None, stack="normalize"),
        )
        .add_selection(y_domain)
    ).interactive()

    histo_x = filter_transform_all(
        alt.Chart(df)
        .properties(height=80, width=500)
        .mark_bar(
            opacity=0.9,
        )
        .encode(
            x=alt.X(
                "intervention_value:Q",
                axis=None,
                bin=alt.Bin(maxbins=bins, minstep=stepsize_x),
                scale=alt.Scale(domain=[x_min, x_max], clamp=True),
            ),
            color=alt.Color("color:N", scale=None, legend=None),
            y=alt.Y("sum(subject_count):Q", axis=None, stack="normalize"),
        )
        .add_selection(x_domain)
    ).interactive()

    legends_color = {}
    for name, key in multi_color_legend_fields.items():
        legends_color[name] = legend(
            df, selections_multi[name], f"{name}", f"{key}:N", f"color_{name}:N"
        ).transform_filter(selection_base)

    legend_black = {}
    for name, key in multi_legend.items():
        legend_black[name] = legend(
            df,
            selections_multi[name],
            f"{name}",
            f"{key}:N",
            color=alt.value("#323232"),
        ).transform_filter(selection_base)

    pk_plot = alt.vconcat(alt.hconcat(scatter + errorbars_y, histo_y))
    age_weight_plot = (
        errorbars_age + errorbars_weight + scatter.encode(x=x_age, y=y_weight)
    )
    legends_color_plot = alt.vconcat(*legends_color.values())
    legend_black_plot = alt.vconcat(*legend_black.values())

    chart = alt.vconcat(
        histo_x,
        pk_plot & age_weight_plot & bars
        | alt.hconcat(legends_color_plot, legend_black_plot),
    )

    if create_json:
        chart.save(
            f"{path}.json",
        )
    else:
        chart.save(
            f"{path}.html", webdriver="firefox", embed_options={"renderer": "svg"}
        )


def pkdata_by_measurement_type(
    pkdata,
    plotting_categories: PlottingParameter,
    intervention_substances,
    output_substances,
    exclude_study_names,
):

    # filter pkdata for given attributes
    data_dict = {}
    for plotting_category in plotting_categories:
        data = pkdata.filter_intervention(
            f_dosing_in, substances=intervention_substances
        ).filter_output(
            f_mt_in_substance_in,
            measurement_types=plotting_category.measurement_types,
            substances=output_substances,
        )
        data = data.exclude_output(lambda d: d["unit"].isin(plotting_category.units_rm))
        data = data.exclude_intervention(
            lambda d: d["study_name"].isin(exclude_study_names)
        )

        data.outputs["measurement_type"] = plotting_category.joined_measurement_type
        data_dict[plotting_category.joined_measurement_type] = data.copy()
    return data_dict


def _get_pc(measurement_type, plotting_categories):
    for plotting_category in plotting_categories:
        if plotting_category.joined_measurement_type == measurement_type:
            return plotting_category
    return


def results(
    data_dict,
    intervention_substances,
    additional_information,
    url,
    plotting_categories,
    replacements,
):
    # creates one dataframe from PKData instance.
    # infers additional results from body weights.
    results_dict = {}
    for measurement_type, pkd in data_dict.items():
        meta_analysis = MetaAnalysis(pkd, intervention_substances, url)
        meta_analysis.create_results()
        for key, additional_function in additional_information.items():
            meta_analysis.results[key] = meta_analysis.results.apply(
                additional_function, axis=1
            )
        pc = _get_pc(measurement_type, plotting_categories)
        meta_analysis.infer_from_body_weight(
            by_intervention=pc.infer_by_intervention, by_output=pc.infer_by_output
        )
        meta_analysis.add_extra_info(replacements)
        results = meta_analysis.results
        results_dict[measurement_type] = results
        print(measurement_type)
        print(len(results))

    return results_dict


def check_legends(df, legend_keys):
    for legend_key in legend_keys:
        assert legend_key in df, f"{legend_key} is not in your data"


def create_navigation_file(results_dict, path):
    navigation = []

    for measurement_type, result_infer in results_dict.items():
        this_nav = {"name": measurement_type, "link": "/#", "dropdown": []}
        for group, df in result_infer.groupby("unit_category"):

            u_unit = ureg(df["unit"].unique()[0])
            u_unit_intervention = ureg(df["intervention_unit"].unique()[0])
            this_dropdown_item = {
                "name": f"{measurement_type} [{u_unit.u :~P}] / dosing [{u_unit_intervention.u :~P}]".capitalize(),
                "link": f"/_pages/{measurement_type}_{group}/",
            }
            this_nav["dropdown"].append(this_dropdown_item)

        navigation.append(this_nav)
        path_data = path / "_data"
        path_data.mkdir(exist_ok=True)
        with open(path / "_data" / "navigation.yml", "w") as f:
            f.write(yaml.dump(navigation))


def create_pages(results_dict, path):
    for measurement_type, result_infer in results_dict.items():
        for group, df in result_infer.groupby("unit_category"):

            u_unit = ureg(df["unit"].unique()[0])
            u_unit_intervention = ureg(df["intervention_unit"].unique()[0])
            intervention_substance = df["intervention_substance"].unique()[0]

            content = {
                "title": f"{measurement_type} [{u_unit.u :~P}] / dosing [{u_unit_intervention.u :~P}]".capitalize(),
                "subtitle": f"Meta analysis of {intervention_substance} {measurement_type}",
                "layout": "pk",
                "hero_height": "80px",
                "json": f"{measurement_type}_{group}.json",
            }
            path_pages = path / "_pages"
            path_pages.mkdir(exist_ok=True)
            with open(path / "_pages" / f"{measurement_type}_{group}.md", "w") as f:
                f.write("---\n")
                f.write(yaml.dump(content))
                f.write("---\n")


def create_plots(
    results_dict, path, multi_legend, multi_color_legend, tooltip, create_json
):
    for measurement_type, result_infer in results_dict.items():
        for group, df in result_infer.groupby("unit_category"):
            path_reports = path / "_static" / "reports"
            path_reports.mkdir(parents=True, exist_ok=True)
            file_name = path_reports / f"{measurement_type}_{group}"
            multi_color_legend_fields = [mc.field for mc in multi_color_legend.values()]
            check_legends(df, [*multi_color_legend_fields, *multi_legend.values()])
            create_interactive_plot(
                df,
                file_name,
                tooltip=tooltip,
                multi_legend=multi_legend,
                multi_color_legend=multi_color_legend,
                create_json=create_json,
            )


def copy_dir(src, dst, substance):
    """
    Recursively copies the content of the directory src to the directory dst.
    If dst doesn't exist, it is created, together with all missing parent directories.
    If a file from src already exists in dst, the file in dst is overwritten.
    Files already existing in dst which don't exist in src are preserved.
    Symlinks inside src are copied as symlinks, they are not resolved before copying.

    :param src:
    :param dst:
    :return:
    """
    dst.mkdir(parents=True, exist_ok=True)
    for item in os.listdir(src):
        s = src / item
        d = dst / item
        if s.is_dir():
            copy_dir(s, d, substance)
        else:
            if str(s).endswith(".md") or str(s).endswith(".yml"):
                with open(s, "r") as f:
                    data = f.read()
                    with open(d, "w") as df:
                        df.write(
                            data.replace("?Substance?", substance.capitalize()).replace(
                                "?substance?", substance
                            )
                        )
            else:
                shutil.copy2(str(s), str(d))


def interactive_plot_factory(
    pkdata,
    plotting_categories,
    intervention_substances,
    output_substances,
    exclude_study_names,
    multi_color_legend,
    multi_legend,
    additional_information,
    tooltip,
    path,
    url="http://0.0.0.0:8081",
    create_json=True,
    replacements={},
):

    data_dict = pkdata_by_measurement_type(
        pkdata,
        plotting_categories,
        intervention_substances,
        output_substances,
        exclude_study_names,
    )
    results_dict = results(
        data_dict=data_dict,
        intervention_substances=intervention_substances,
        additional_information=additional_information,
        url=url,
        plotting_categories=plotting_categories,
        replacements=replacements,
    )
    copy_dir(Path(__file__).parent / "template", path, "_".join(output_substances))
    create_navigation_file(results_dict, path)
    create_pages(results_dict, path)
    create_plots(
        results_dict,
        path,
        multi_legend,
        multi_color_legend,
        tooltip,
        create_json=create_json,
    )



"""
Access to important environment variables.
"""
import logging
import os
from urllib import parse as urlparse

from pkdb_analysis.logging_utils import bcolors


logger = logging.getLogger(__name__)

try:
    API_BASE = os.environ["API_BASE"]
    # fix terminal slash
    if API_BASE.endswith("/"):
        API_BASE = API_BASE[:-1]

    USER = os.environ["USER"]
    PASSWORD = os.environ["PASSWORD"]

    API_URL = API_BASE + "/api/v1"

except KeyError:

    USER = None
    PASSWORD = None
    API_BASE = "https://alpha.pk-db.com"

    API_URL = API_BASE + "/api/v1"
    logger.warning(
        f"Environment variables have not been initialized. "
        f"1. add authentication credentials; and 2. run {bcolors.OKBLUE}set -a && "
        f"source .env.local{bcolors.ENDC}. "
        f"Queries will be performed as a anonymous user."
    )



"""
Helpers related to logging
"""
import coloredlogs


coloredlogs.install(
    level="INFO",
    fmt="%(levelname)s %(message)s",  # "%(module)s:%(lineno)s %(funcName)s %(levelname) -10s %(message)s"
)


class bcolors:
    """ Colors for styling log. """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BGWHITE = "\033[47m"
    BGBLACK = "\033[49m"
    WHITE = "\033[37m"
    BLACK = "\033[30m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"



import pytest

from pkdb_analysis.reports import create_table_report
from pkdb_analysis.test import TEST_HDF5


# FIXME
@pytest.mark.skip("FIXME")
def test_tables(tmp_path):
    output_path = tmp_path / "tables"
    xlsx_path = output_path / "tables.xlsx"
    create_table_report(
        h5_data_path=TEST_HDF5,
        dosing_substances=["torasemide"],
        report_substances=[
            "torasemide",
            "torasemide-M1",
            "torasemide-M3",
            "torasemide-M5",
        ],
        excel_path=xlsx_path,
        # google_sheets="TorasemideTables",
        tsv_path=output_path,
        query_data=False,
    )

    assert xlsx_path.exists()
    assert (output_path / "studies.tsv").exists()
    assert (output_path / "pharmacokinetics.tsv").exists()
    assert (output_path / "timecourses.tsv").exists()



"""
Definition of data and files for the tests.
The files are located in the data directory.
"""
from pathlib import Path

TEST_PATH = Path(__file__).parent  # directory of test files
TESTDATA_PATH = TEST_PATH / "data"  # directory of data for tests
TEST_HDF5 = TESTDATA_PATH / "test.h5"



""" Module for analysis of PKData"""
import os

import matplotlib.pyplot as plt
import numpy as np
import pint
from matplotlib import ticker
from matplotlib.lines import Line2D
from matplotlib.ticker import FormatStrFormatter


ureg = pint.UnitRegistry()

# ---- Styles for plotting ----

import matplotlib.font_manager as font_manager


font = font_manager.FontProperties(
    family="Roboto Mono",
    weight="normal",
    style="normal",
    size=16,
)
plt.rcParams.update(
    {
        "axes.labelsize": "20",
        "axes.labelweight": "bold",
        "axes.titlesize": "medium",
        "axes.titleweight": "bold",
        "legend.fontsize": "20",
        "xtick.labelsize": "20",
        "ytick.labelsize": "20",
        "figure.facecolor": "1.00",
    }
)
# ------------------------------


def _get_one(d):
    d = d.dropna()
    assert len(set(d)) <= 1, set(d)
    return list(set(d))[0]


def _str_all(d):
    d = d.dropna()
    return ", ".join(set(d))


def mscatter(x, y, ax=None, m=None, **kw):
    import matplotlib.markers as mmarkers

    ax = ax or plt.gca()
    sc = ax.scatter(x, y, **kw)
    if (m is not None) and (len(m) == len(x)):
        paths = []
        for marker in m:
            if isinstance(marker, mmarkers.MarkerStyle):
                marker_obj = marker
            else:
                marker_obj = mmarkers.MarkerStyle(marker)
            path = marker_obj.get_path().transformed(marker_obj.get_transform())
            paths.append(path)
        sc.set_paths(paths)
    return sc


def figure_category(d):
    if d["per_bw"] and d["intervention_per_bw"]:
        return "rel_output_rel_intervention"
    elif d["per_bw"] and not d["intervention_per_bw"]:
        return "rel_output_abs_intervention"
    elif not d["per_bw"] and not d["intervention_per_bw"]:
        return "abs_output_abs_intervention"
    elif not d["per_bw"] and d["intervention_per_bw"]:
        return "abs_output_rel_intervention"


def create_plots(
    data,
    fig_path,
    color_by=None,
    color_mapping=None,
    nrows=2,
    ncols=2,
    figsize=(30, 30),
    log_y=False,
    formats=["png", "svg"],
):
    data["plotting_category"] = data[["per_bw", "intervention_per_bw"]].apply(
        figure_category, axis=1
    )
    figure, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
    axes_iter = iter(axes.flatten())
    substance = _get_one(data.substance)
    substance_intervention = _str_all(data.substance_intervention)
    measurement_type = _get_one(data.measurement_type)

    for plotting_category, data_category in data.groupby("plotting_category"):

        unit = _get_one(data_category.unit)
        u_unit = ureg(unit)
        unit_intervention = _get_one(data_category.unit_intervention)
        u_unit_intervention = ureg(unit_intervention)
        ax = next(axes_iter)

        y_axis_label = f"{substance} {measurement_type}"

        ax.set_ylabel(f"{y_axis_label} [{u_unit.u :~P}]")
        x_label = "$Dose_{" + substance_intervention + "}$"
        ax.set_xlabel(f"{x_label} [{u_unit_intervention.u :~P}]")
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))

        max_values = np.array(
            [
                data_category["value"].max(),
                data_category["mean"].max(),
                data_category["median"].max(),
            ]
        )
        max_values = max_values[~np.isnan(max_values)]
        df_subplot_max = np.max(max_values) * 1.05
        df_figure_x_max = data_category.value_intervention.max() * 1.05
        df_figure_min = 0

        individuals_data = data_category[data_category.group_pk == -1]
        x = individuals_data.value_intervention
        y = individuals_data.value
        color = list(individuals_data.color)
        marker = list(individuals_data.marker)
        mscatter(x, y, ax=ax, color=color, m=marker, alpha=0.7, label=None, s=20)

        group_data = data_category[data_category.group_pk != -1]

        x_group_max = group_data["value_intervention"].max()
        y_group_max = group_data["mean"].max()
        for group, df_group in group_data.iterrows():
            x_group = df_group["value_intervention"]

            if np.isnan(df_group["mean"]):
                y_group = df_group["median"]

            else:
                y_group = df_group["mean"]
            yerr_group = df_group.se
            ms = df_group.group_count + 5
            color = df_group.color
            marker = df_group.marker

            ax.errorbar(
                x_group,
                y_group,
                yerr=yerr_group,
                xerr=0,
                color=color,
                fmt=marker,
                ms=ms,
                alpha=0.7,
            )

            txt = df_group["study_name"]

            data_values = [
                x_group + (0.01 * x_group_max),
                y_group + (0.01 * y_group_max),
            ]
            isnan = np.isnan(np.array(data_values))
            if not any(isnan):
                ax.annotate(
                    txt,
                    (x_group + (0.01 * x_group_max), y_group + (0.01 * y_group_max)),
                    alpha=0.7,
                )

        legend_elements = []

        for plotting_type, d in data_category.groupby("plotting_type"):
            individuals_data = d[d.group_pk == -1]
            group_data = d[d.group_pk != -1]

            individuals_number = len(individuals_data)
            group_number = len(group_data)
            total_group_individuals = group_data["group_count"].sum()
            color = _get_one(d.color)

            label_text = f"{plotting_type:<10} I: {individuals_number:<3} G: {group_number:<2} TI: {int(total_group_individuals + individuals_number):<3}"
            label = Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                label=label_text,
                markerfacecolor=color,
                markersize=10,
            )
            legend_elements.append(label)

        legend2_elements = [
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                label="PK DATA",
                markerfacecolor="black",
                markersize=10,
            ),
            Line2D(
                [0],
                [0],
                marker="s",
                color="w",
                label="PK FROM TIME COURSE",
                markerfacecolor="black",
                markersize=10,
            ),
            Line2D(
                [0],
                [0],
                marker="v",
                color="w",
                label="PK FROM BODY WEIGHT",
                markerfacecolor="black",
                markersize=10,
            ),
        ]

        legend3_elements = [
            Line2D(
                [0],
                [10],
                marker="o",
                color="w",
                label="1",
                markerfacecolor="black",
                markersize=5 + 1,
            ),
            Line2D(
                [0],
                [20],
                marker="o",
                color="w",
                label="10",
                markerfacecolor="black",
                markersize=5 + 10,
            ),
            Line2D(
                [0],
                [50],
                marker="o",
                color="w",
                label="30",
                markerfacecolor="black",
                markersize=5 + 30,
            ),
        ]

        ax.set_title(f"{plotting_category}")
        ax.set_xlim(left=0, right=df_figure_x_max)
        # ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        ax.yaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))

        leg1 = ax.legend(handles=legend_elements, prop=font, loc="upper right")

        leg2 = ax.legend(handles=legend2_elements, prop=font, loc="upper left")
        leg3 = ax.legend(
            handles=legend3_elements, prop=font, labelspacing=1.3, loc="center right"
        )
        leg3.set_title(title="GROUP SIZE", prop=font)
        ax.add_artist(leg2)
        ax.add_artist(leg1)

        if log_y:
            ax.set_yscale("log")
            ax.set_ylim(bottom=df_figure_min, top=df_subplot_max)
        else:
            ax.set_ylim(bottom=0, top=df_subplot_max)
    for format in formats:
        figure.savefig(
            os.path.join(fig_path, f"{measurement_type}.{format}"),
            bbox_inches="tight",
            dpi=72,
            format=format,
        )
    # figure.savefig(os.path.join(fig_path, f"{measurement_type}.png"), bbox_inches="tight", dpi=72)



from pkdb_analysis import PKFilter, PKFilterFactory


def test_empty_filter():
    filter = PKFilter(normed=None)

    assert len(filter.groups) == 0
    assert len(filter.individuals) == 0
    assert len(filter.interventions) == 0
    assert len(filter.outputs) == 0
    assert len(filter.timecourses) == 0

    d = filter.to_dict()
    assert isinstance(d, dict)
    for filter_key in PKFilter.KEYS:
        assert filter_key in d


def test_study_sid_filter():
    filter = PKFilterFactory.by_study_sid("PKDB00181")
    assert isinstance(filter, PKFilter)
    d = filter.to_dict()
    for filter_key in PKFilter.KEYS:
        assert filter_key in d
        assert "study_sid" in d[filter_key]
        assert d[filter_key]["study_sid"] == "PKDB00181"


def test_study_name_filter():
    filter = PKFilterFactory.by_study_name("Allonen1981")
    assert isinstance(filter, PKFilter)
    d = filter.to_dict()
    for filter_key in PKFilter.KEYS:
        assert filter_key in d
        assert "study_name" in d[filter_key]
        assert d[filter_key]["study_name"] == "Allonen1981"
